import ast
import sys
import time
from collections import namedtuple
from contextlib import contextmanager
from contextvars import ContextVar
from itertools import count

from varname import ImproperUseError, VarnameRetrievingError, argname, varname
from varname.utils import get_node

global_context = ContextVar("global_context", default=())
global_inherited = ContextVar("global_inherited", default={})

_block_classes = {
    ast.If: ("body", "orelse"),
    ast.For: ("body", "orelse"),
    ast.While: ("body", "orelse"),
    ast.FunctionDef: ("body",),
    ast.AsyncFunctionDef: ("body",),
    ast.With: ("body",),
    ast.AsyncWith: ("body",),
    ast.AsyncFor: ("body", "orelse"),
}


_improper_nullary_give_error = (
    "give() with no arguments must immediately follow an assignment"
)


special_keys = {}


global_count = count(0)


def register_special(key):
    """Return a decorator to register a function for a special key.

    The function is called with no arguments whenever the special key is
    requested, e.g. with ``Giver(special=["$specialkey"])``.

    Use ``sys._getframe(3)`` to get the frame in which give() was called.

    Example:
        .. code-block:: python

            @register_special("$time")
            def _special_time():
                return time.time()

    Arguments:
        key: The key, conventionally starting with a "$".
    """

    def deco(func):
        special_keys[key] = func
        return func

    return deco


@register_special("$time")
def _special_time():
    return time.time()


@register_special("$frame")
def _special_frame():
    return sys._getframe(3)


LinePosition = namedtuple("LinePosition", ["name", "filename", "lineno"])


@register_special("$line")
def _special_line():
    fr = sys._getframe(3)
    co = fr.f_code
    return LinePosition(co.co_name, co.co_filename, fr.f_lineno)


def _find_targets(target):
    if isinstance(target, ast.Tuple):
        results = []
        for t in target.elts:
            results += _find_targets(t)
        return results
    else:
        return [target.id]


def _find_above(frame):
    node = get_node(frame + 1)
    if node is None:
        raise VarnameRetrievingError(
            "Cannot retrieve the node where the function is called"
        )

    while node.parent is not None:
        parent = node.parent
        fields = _block_classes.get(type(parent), None)
        if fields is None:
            node = parent
            continue
        else:
            for field in fields:
                f = getattr(parent, field)
                if node in f:
                    idx = f.index(node)
                    if idx == 0:
                        raise ImproperUseError(_improper_nullary_give_error)

                    assignment = f[idx - 1]

                    if isinstance(assignment, ast.Assign):
                        target = assignment.targets[-1]
                        names = _find_targets(target)
                    elif isinstance(assignment, (ast.AugAssign, ast.AnnAssign)):
                        names = [assignment.target.id]
                    else:
                        raise ImproperUseError(_improper_nullary_give_error)

                    fr = sys._getframe(frame)
                    rval = {}

                    for name in names:
                        if name in fr.f_locals:
                            rval[name] = fr.f_locals[name]
                        elif name in fr.f_globals:
                            rval[name] = fr.f_globals[name]
                        else:  # pragma: no cover
                            # I am not sure how to trigger this
                            raise Exception("Could not resolve value")
                    return rval

            else:  # pragma: no cover
                # I am not sure how to trigger this
                raise Exception("Could not find node position")

    # I am not sure how to trigger this
    raise Exception("Could not find node")  # pragma: no cover


def resolve(frame, func, args):
    """Return a {variable_name: value} dictionary depending on usage.

    * ``len(args) == 0`` => Use the variable assigned in the line before the call.
    * ``len(args) == 1`` => Use the variable the call is assigned to.
    * ``len(args) >= 1`` => Use the variables passed as arguments to the call.

    Arguments:
        frame: The number of frames to go up to find the context.
        func: The Giver object that was called.
        args: The arguments given to the Giver.
    """
    nargs = len(args)

    if nargs == 0:
        return _find_above(frame=frame + 2)

    if nargs == 1:
        try:
            assigned_to = varname(frame=frame + 1, strict=True, raise_exc=False)
        except ImproperUseError:
            assigned_to = None
        if assigned_to is not None:
            return {assigned_to: args[0]}

    argnames = argname("args", func=func, frame=frame + 1, vars_only=False)
    if argnames is None:  # pragma: no cover
        # I am not sure how to trigger this
        raise Exception("Could not resolve arg names")

    return {name: value for name, value in zip(argnames, args)}


class Giver:
    """Giver of key/value pairs.

    ``Giver`` is the class of the ``give`` object.

    Arguments:
        keys:
            List of default keys to give. If ``keys=["x"]``, then
            ``self(123)`` will give ``{"x": 123}``.
        special:
            List of special keys to give (e.g. "$line", "$time", etc.)
        extra:
            Extra key/value pairs to give.
        context:
            The ContextVar that contains a list of handlers to call
            when something is given.
        inherited:
            A ContextVar to use for inherited key/value pairs to give,
            as set by ``with self.inherit(key=value): ...``.
        transform:
            A function from dict to dict that modifies the values to
            give.
    """

    def __init__(
        self,
        *,
        keys=None,
        special=[],
        extra={},
        context=global_context,
        inherited=global_inherited,
        transform=None,
    ):
        self.keys = keys
        self.special = special
        self.extra = extra
        self.context = context
        self.inherited = inherited
        self.transform = transform

    def copy(
        self,
        keys=None,
        special=None,
        extra=None,
        context=None,
        inherited=None,
        transform=None,
    ):
        """Copy this Giver with modified parameters."""
        return type(self)(
            keys=self.keys if keys is None else keys,
            special=self.special if special is None else special,
            extra=self.extra if extra is None else extra,
            context=self.context if context is None else context,
            inherited=self.inherited if inherited is None else inherited,
            transform=self.transform if transform is None else transform,
        )

    @property
    def line(self):
        """Return a giver that gives the line where it is called."""
        return self.copy(special=(*self.special, "$line"))

    @property
    def time(self):
        """Return a giver that gives the time where it is called."""
        return self.copy(special=(*self.special, "$time"))

    @contextmanager
    def inherit(self, **keys):
        """Create a context manager within which extra values are given.

        .. code-block:: python

            with give.inherit(a=1):
                give(b=2)   # gives {"a": 1, "b": 2}

        Arguments:
            keys: The key/value pairs to give within the block.
        """
        inh = self.inherited.get()
        token = self.inherited.set({**inh, **keys})
        try:
            yield
        finally:
            self.inherited.reset(token)

    @contextmanager
    def wrap(self, name, **keys):
        """Create a context manager that marks the beginning/end of the block.

        ``wrap`` first creates a unique ID to identify the block,
        then gives the ``$wrap`` sentinel with name, uid and step="begin"
        at the beginning of it gives the same ``$wrap`` but with step="end"
        at the end of the block.

        :meth:`giving.gvn.ObservableProxy.wrap` is the corresponding
        method on the ObservableProxy returned by ``given()`` and it
        can be used to wrap another context manager on the same block.
        :meth:`giving.gvn.ObservableProxy.group_wrap` is another method
        that uses the sentinels produced by ``wrap``.

        .. code-block:: python

            with give.wrap("W", x=1):  # gives: {"$wrap": {"name": "W", "step": "begin", "id": ID}, "x": 1}
                ...
            # end block, gives: {"$wrap": {"name": "W", "step": "end", "id": ID}, "x": 1}

        Arguments:
            name: The name to associate to this wrap block.
            keys: Extra key/value pairs to give along with the sentinels.
        """
        num = next(global_count)
        self.produce({"$wrap": {"name": name, "step": "begin", "id": num}, **keys})
        try:
            yield
        finally:
            self.produce({"$wrap": {"name": name, "step": "end", "id": num}, **keys})

    @contextmanager
    def wrap_inherit(self, name, **keys):
        """Shorthand for using wrap and inherit.

        .. code-block:: python

            with give.wrap_inherit("W", a=1):
                ...

        Is equivalent to:

        .. code-block:: python

            with give.inherit(a=1):
                with give.wrap("W"):
                    ...

        Arguments:
            name: The name to associate to this wrap block.
            keys: Key/value pairs to inherit.
        """
        with self.inherit(**keys):
            with self.wrap(name):
                yield

    def produce(self, values):
        """Give the values dictionary."""
        for special in self.special:
            values[special] = special_keys[special]()

        if self.extra:
            values = {**self.extra, **values}

        inh = self.inherited.get()
        if inh is not None:
            values = {**inh, **values}

        for handler in self.context.get():
            handler(values)

    def variant(self, fn):
        """Create a version of give that transforms the data.

        .. code-block:: python

            @give.variant
            def give_image(data):
                return {"image": data}

            ...

            give_image(x, y)  # gives {"image": {"x": x, "y": y}}

        Arguments:
            fn: A function from a dict to a dict.
            give: The base give function to wrap (defaults to global give).
        """
        return self.copy(transform=fn)

    def __call__(self, *args, **values):
        """Give the args and values."""
        h = self.context.get()
        if h:
            if self.keys:
                if len(args) != len(self.keys):
                    raise ImproperUseError(
                        f"Giver for {self.keys} must have {len(self.keys)} positional argument(s)."
                    )
                keyed = dict(zip(self.keys, args))
                values = {**keyed, **values}
            elif args:
                values = {**resolve(1, self, args), **values}
            elif not values:
                values = resolve(1, self, ())

            if self.transform:
                values = self.transform(values)

            self.produce(values)

        if len(args) == 1:
            return args[0]
        else:
            return None


def giver(*keys, **extra):
    """Create a Giver to give the specified keys, plus extra values.

    .. code-block:: python

        g = giver("x", y=1)
        give(3)   # gives {"x": 3, "y": 1}

    """
    normal = [k for k in keys if not k.startswith("$")]
    special = [k for k in keys if k.startswith("$")]
    return Giver(keys=normal, special=special, extra=extra)



from contextvars import ContextVar
from functools import partial
from types import SimpleNamespace

from .gvn import Given
from .gvr import Giver, global_context


def make_give(context=None):
    """Create independent give/given.

    The resulting functions share their own ``ContextVar``, which
    makes them independent from the main instances of ``give`` and
    ``given``.

    Arguments:
        context: The ``ContextVar`` set by ``given`` and used by ``give``,
            or ``None`` if a new ``ContextVar`` is to be created.

    Returns:
        A SimpleNamespace with attributes ``context``, ``give`` and
        ``given``.
    """

    context = context or ContextVar("context", default=())
    give = Giver(context=context)
    given = partial(Given, context=context)

    return SimpleNamespace(
        context=context,
        give=give,
        given=given,
    )


_global_given = make_give(context=global_context)

give = _global_given.give
given = _global_given.given



from . import operators as op
from .api import give, given, make_give
from .extraops import reducer
from .gvn import Failure, Given, ObservableProxy
from .gvr import Giver, giver, global_context, register_special, resolve
from .version import version



import inspect
from contextlib import contextmanager
from functools import partial, wraps

import rx

from . import operators
from .executors import Breakpoint, Displayer
from .gvr import Giver, global_context
from .utils import lax_function


class Failure(Exception):
    """Error type raised by fail() by default."""


class DisposableWrapper:
    """Wraps a Disposable to install a custom context manager."""

    def __init__(self, obs, disposable):
        self.obs = obs
        self.disposable = disposable
        self.dispose = self.disposable.dispose

    def __enter__(self):
        self.obs.__enter__()
        self.disposable.__enter__()

    def __exit__(self, exc_type=None, exc=None, tb=None):
        self.disposable.__exit__(exc_type, exc, tb)
        self.obs.__exit__(exc_type, exc, tb)


class ObservableProxy:
    """Defines a rich interface for an Observable."""

    """Wraps an Observable_ to provide a richer interface.

    .. _Observable: https://rxpy.readthedocs.io/en/latest/reference_observable.html

    For convenience, all operators in :mod:`giving.operators` are
    provided as methods on these objects.
    """

    def __init__(self, obs):
        assert not isinstance(obs, ObservableProxy)
        self.obs = obs

    def _copy(self, new_obs):
        """Copy this observable with a new underlying observable."""
        return type(self)(new_obs)

    ###################
    # Wrapped methods #
    ###################

    def pipe(self, *args, **kwargs):
        """Pipe one or more operators.

        Returns: An ObservableProxy.
        """
        return self._copy(self.obs.pipe(*args, **kwargs))

    def subscribe(self, *args, **kwargs):
        """Subscribe a function to this Observable stream.

        .. code-block:: python

            with given() as gv:
                gv.subscribe(print)

                results = []
                gv["x"].subscribe(results.append)

                give(x=1)  # prints {"x": 1}
                give(x=2)  # prints {"x": 2}

                assert results == [1, 2]

        Arguments:
            observer:
                The object that is to receive notifications.
            on_error:
                Action to invoke upon exceptional termination of the
                observable sequence.
            on_completed:
                 Action to invoke upon graceful termination of the
                 observable sequence.
            on_next:
                Action to invoke for each element in the observable
                sequence.

        Returns:
            An object representing the subscription with a ``dispose()``
            method to remove it.
        """
        disposable = self.obs.subscribe(*args, **kwargs)
        return DisposableWrapper(self, disposable)

    def subscribe_(self, *args, **kwargs):
        disposable = self.obs.subscribe_(*args, **kwargs)
        return DisposableWrapper(self, disposable)

    #########################
    # Special subscriptions #
    #########################

    def accum(self, obj=None):
        """Accumulate into a list or set.

        Arguments:
            obj: The object in which to accumulate, either a list or
                a set. If not provided, a new list is created.

        Returns:
            The object in which the values will be accumulated.
        """
        if obj is None:
            obj = []
            method = obj.append
        elif isinstance(obj, list):
            method = obj.append
        elif isinstance(obj, set):
            method = obj.add
        else:
            raise TypeError("Argument to accum must be a list or set.")

        self.subscribe(method)
        return obj

    def breakpoint(self, **kwargs):  # pragma: no cover
        """Trigger a breakpoint on every entry.

        Arguments:
            skip:
                A list of globs corresponding to modules to skip during debugging,
                for example ``skip=["giving.*"]`` would skip all frames that are in
                the ``giving`` module.
        """
        return self.subscribe(Breakpoint(**kwargs))

    def breakword(self, **kwargs):  # pragma: no cover
        """Trigger a breakpoint using ``breakword``.

        This feature requires the ``breakword`` package to be installed, and the
        :func:`~giving.operators.tag` operator to be applied.

        .. code-block:: python

            gvt = gv.tag()
            gvt.display()
            gvt.breakword()

        The above will display words next to each entry. Set the BREAKWORD environment
        to one of these words to set a breakpoint when it is printed.

        Arguments:
            skip:
                A list of globs corresponding to modules to skip during debugging,
                for example ``skip=["giving.*"]`` would skip all frames that are in
                the ``giving`` module.
            word:
                Only trigger the breakpoint on the given word.
        """
        return self.subscribe(Breakpoint(use_breakword=True, **kwargs))

    def display(self, *, breakword=None, skip=[], **kwargs):
        """Pretty-print each element.

        Arguments:
            colors:
                Whether to colorize the output or not.
            time_format:
                How to format the time (if present), e.g. ``"%Y-%m-%d %H:%M:%S"``
            breakword:
                If not None, run ``self.breakword(word=breakword)``.
            skip:
                If ``breakword is not None``, pass ``skip`` to the debugger.
        """
        sub = self.subscribe(Displayer(**kwargs))
        if breakword:  # pragma: no cover
            self.breakword(word=breakword, skip=skip)
        return sub

    def fail(self, exc_type=Failure):
        """Raise an exception if the stream produces anything.

        Arguments:
            exc_type:
                The exception type to raise. Will be passed the next data
                element, and the result is raised. Defaults to
                :class:`~giving.gvn.Failure`.
        """

        def _fail(data):
            raise exc_type(data)

        return self.subscribe(_fail)

    def fail_if_empty(self, exc_type=Failure):
        """Raise an exception if the stream is empty.

        Arguments:
            exc_type:
                The exception type to raise. Defaults to :class:`~giving.gvn.Failure`.
        """

        def _fail(is_empty):
            if is_empty:
                raise exc_type(is_empty)

        return self.is_empty().subscribe(_fail)

    def give(self, *keys, **extra):
        """Give each element.

        This calls :func:`~giving.api.give` for each value in the stream.

        Be careful using this method because it could easily lead to an infinite loop.

        Arguments:
            keys: Key(s) under which to give the elements.
            extra: Extra key/value pairs to give along with the rest.
        """
        giver = Giver(keys=keys, extra=extra)

        if len(keys) == 0:
            gv = giver.produce
        elif len(keys) == 1:
            gv = giver
        else:

            def gv(x):
                return giver(*x)

        return self.subscribe(gv)

    def ksubscribe(self, fn):
        """Subscribe a function called with keyword arguments.

        .. note::
            The function passed to ``ksubscribe`` is wrapped with
            :func:`~giving.utils.lax_function`, so it is not necessary
            to add a ``**kwargs`` argument for keys that you do not need.

        .. code-block:: python

            gv.ksubscribe(lambda x, y=None, z=None: print(x, y, z))
            give(x=1, z=2, abc=3)  # Prints 1, None, 2

        Arguments:
            fn: The function to call.
        """
        fn = lax_function(fn)
        return self.subscribe(lambda data: fn(**data))

    def kwrap(self, name, fn=None, return_function=False):
        """Subscribe a context manager, corresponding to :meth:`~giving.gvr.Giver.wrap`.

        ``obs.kwrap(fn)`` is shorthand for ``obs.wrap(fn, pass_keys=True)``.

        .. note::
            The function passed to ``ksubscribe`` is wrapped with
            :func:`~giving.utils.lax_function`, so it is not necessary
            to add a ``**kwargs`` argument for keys that you do not need.

        .. code-block:: python

            @gv.kwrap
            def _(x):
                print(">", x)
                yield
                print("<", x)

            with give.wrap(x=1):      # prints >1
                ...
                with give.wrap(x=2):  # prints >2
                    ...
                ...                   # prints <2
            ...                       # prints <1

        Arguments:
            name: The name of the wrap block to subscribe to.
            fn: The wrapper function. The arguments to ``give.wrap`` are transferred to
                this function as keyword arguments.
        """
        return self.wrap(name, fn, pass_keys=True, return_function=return_function)

    def print(self, format=None, skip_missing=False):
        """Print each element of the stream.

        Arguments:
            format: A format string as would be used with ``str.format``.
            skip_missing: Whether to ignore KeyErrors due to missing entries in the format.
        """
        obs = self
        if format is not None:
            obs = self.format(format, skip_missing=skip_missing)
        return obs.subscribe(print)

    def wrap(self, name, fn=None, pass_keys=False, return_function=False):
        """Subscribe a context manager, corresponding to :meth:`~giving.gvr.Giver.wrap`.

        .. code-block:: python

            @gv.wrap("main")
            def _():
                print("<")
                yield
                print(">")

            with give.wrap("main"):    # prints <
                ...
                with give.wrap("sub"):
                    ...
                ...
            ...                        # prints >

        Arguments:
            name: The name of the wrap block to subscribe to.
            fn: The wrapper function OR an object with an ``__enter__`` method. If the wrapper is
                a generator, it will be wrapped with ``contextmanager(fn)``.
                If a function, it will be called with no arguments, or with the arguments given to
                ``give.wrap`` if ``pass_keys=True``.
            pass_keys: Whether to pass the arguments to ``give.wrap`` to this function as keyword
                arguments. You may use :meth:`kwrap` as a shortcut to ``pass_keys=True``.
        """
        if fn is None:
            return partial(self.wrap, name, pass_keys=pass_keys, return_function=True)

        if pass_keys:
            fn = lax_function(fn)

        if inspect.isgeneratorfunction(fn):
            fn = contextmanager(fn)

        managers = {}

        def watch(data):
            wr = data.get("$wrap", None)
            if wr is None or wr["name"] is not name:
                return

            if wr["step"] == "begin":
                key = wr["id"]
                assert key not in managers
                if hasattr(fn, "__enter__"):
                    manager = fn
                elif pass_keys:
                    manager = fn(**data)
                else:
                    manager = fn()

                managers[key] = manager
                manager.__enter__()

            if wr["step"] == "end":
                key = wr["id"]
                managers[key].__exit__(None, None, None)
                del managers[key]

        disposable = self.subscribe(watch)
        if return_function:
            return fn
        else:
            return disposable

    def __or__(self, other):
        """Alias for :func:`~giving.operators.merge`.

        Merge this ObservableProxy with another.
        """
        return self.merge(other)

    __ror__ = __or__

    def __rshift__(self, subscription):
        """Alias for :meth:`subscribe`.

        If ``subscription`` is a list or a set, accumulate into it.
        """
        if isinstance(subscription, list):
            subscription = subscription.append
        elif isinstance(subscription, set):
            subscription = subscription.add
        return self.subscribe(subscription)

    def __getitem__(self, item):
        """Mostly an alias for :meth:`~giving.operators.getitem`.

        Extra feature: if the item starts with ``"?"``, ``getitem`` is called with
        ``strict=False``.
        """
        if not isinstance(item, tuple):
            item = (item,)
        strict = not any("?" in x for x in item if isinstance(x, str))
        item = [x.lstrip("?") if isinstance(x, str) else x for x in item]
        return self.getitem(*item, strict=strict)

    #################################################
    # Methods for each operator in giving.operators #
    #################################################

    ...  # The methods are set with _put_opmethods below


def _opmethod(operator):
    @wraps(operator)
    def f(self, *args, **kwargs):
        return self.pipe(operator(*args, **kwargs))

    return f


def _put_opmethods(cls, names):
    for name in names:
        operator = getattr(operators, name)
        setattr(cls, name, _opmethod(operator))


_operations = [op for op in operators.__all__ if op not in {"pipe"}]
_put_opmethods(ObservableProxy, _operations)


class Given(ObservableProxy):
    """Observable that streams the values given using ``give()``.

    Instances of Given must be activated as a context manager in order to
    work, and each instance can only be activated once.

    .. code-block:: python

        # Instantiate a Given and set up a pipeline of operations
        gv = given()
        gv["?x"].max().print("max(x) = {}")

        # Calls to give() do nothing yet
        give(x=12345)

        # We activate gv using a ``with`` statement:
        with gv:
            # NOW, calls to give() are fed into the pipeline
            give(x=1, y=1)

            # You can still add to the pipeline, but the new operations
            # will only pick up future calls to give()
            gv["?y"].min().print("min(y) = {}")

            give(x=2, y=2)
            ...

        # min, max and other reductions are resolved after the block ends
        # Prints max(x) = 2 (x=12345 is outside the block and ignored)
        # and min(y) = 2 (y=1 is before the call to min()).

        with gv:
            # ERROR because gv can only be activated once
            ...

    Arguments:
        context:
            The ContextVar to use to sync with ``give``.
    """

    def __init__(self, context=global_context, _obs=None, _root=None):
        self._token = None
        self._context = context

        if _obs is None:
            assert _root is None
            _root = self

            self._observers = []

            def make(observer, scheduler):
                self._observers.append(observer)

            _obs = rx.create(make)

        self._root = _root
        super().__init__(_obs)

    #################
    # Extra methods #
    #################

    @contextmanager
    def values(self):
        """Context manager to accumulate the stream into a list.

        .. code-block:: python

            with given()["?x"].values() as results:
                give(x=1)
                give(x=2)

            assert results == [1, 2]

        Note that this will activate the root given() including all subscriptions
        that it has (directly or indirectly).
        """
        with self._root:
            results = self.accum()
            yield results

    def exec(self, fn, *args, **kwargs):
        """Run a function in the context of this Given.

        .. code-block:: python

            def main():
                give(x=1)
                give(x=2)

            gv = given()
            gv["x"].print()
            gv.exec(main)  # prints 1, 2

        Arguments:
            fn: The function to run.
            args: Positional arguments to pass to fn.
            kwargs: Keyword arguments to pass to fn.
        """
        with self._root:
            fn(*args, **kwargs)

    def eval(self, fn, *args, **kwargs):
        """Run a function in the context of this Given and get the values.

        .. code-block:: python

            def main():
                give(x=1)
                give(x=2)

            values = given()["x"].eval(main)
            assert values == [1, 2]

        Arguments:
            fn: The function to run.
            args: Positional arguments to pass to fn.
            kwargs: Keyword arguments to pass to fn.
        """
        with self.values() as results:
            fn(*args, **kwargs)
        return results

    #################
    # Instantiation #
    #################

    def _copy(self, new_obs):
        """Copy this observable with a new underlying observable."""
        return type(self)(context=None, _obs=new_obs, _root=self._root)

    def _push(self, data):
        """Push data to the stream."""
        for obs in self._observers:
            obs.on_next(data)

    def __enter__(self):
        if self._root is not self:
            self._root.__enter__()
            return self

        if self._token is not None:
            raise Exception("An instance of given() can only be entered once")

        h = self._context.get()
        self._token = self._context.set((*h, self._push))
        return self

    def __exit__(self, exc_type=None, exc=None, tb=None):
        if self._root is not self:
            self._root.__exit__(exc_type, exc, tb)
            return

        for obs in self._observers:
            obs.on_completed()
        self._context.reset(self._token)
        self._observers.clear()
        self._token = True



"""Extra operators for giving."""

import builtins
import operator
from collections import deque
from types import FunctionType

import rx
from rx import operators as rxop
from rx.operators import NotSet

from .utils import lax_function, reducer


def roll(n, reduce=None, seed=NotSet):  # noqa: F811
    """Group the last n elements, giving a sequence of overlapping sequences.

    For example, this can be used to compute a rolling average of the 100 last
    elements (however, ``average(scan=100)`` is better optimized).

    .. code-block:: python

        op.roll(100, lambda xs: sum(xs) / len(xs))

    .. marble::
        :alt: roll

        --1--2--3---4---5---6---|
        [         roll(3)       ]
        --1--12-123-234-345-456-|

    Arguments:
        n: The number of elements to group together.
        reduce: A function to reduce the group.

            It should take five arguments:

                * **last**: The last result.
                * **add**: The element that was just added. It is the last element
                  in the elements list.
                * **drop**:
                  The element that was dropped to make room for the
                  added one. It is *not* in the elements argument.
                  If the list of elements is not yet of size n, there is
                  no need to drop anything and drop is None.
                * **last_size**: The window size on the last invocation.
                * **current_size**: The window size on this invocation.

            Defaults to returning the deque of elements directly.

            .. note::
                The same reference is returned each time in order to save memory, so it
                should be processed immediately.
        seed: The first element of the reduction.
    """
    q = deque(maxlen=n)

    if reduce is not None:

        def queue(current, x):
            drop = q[0] if len(q) == n else NotSet
            last_size = len(q)
            q.append(x)
            current_size = len(q)
            return reduce(
                current,
                x,
                drop=drop,
                last_size=last_size,
                current_size=current_size,
            )

        scan_command = rxop.scan(queue, seed)

    else:

        def queue(q, x):
            q.append(x)
            return q

        scan_command = rxop.scan(queue, q)

    return rxop.pipe(scan_command, rxop.share())


def affix(**streams):
    """Affix streams as extra keys on an existing stream of dicts.

    The affixed streams should have the same length as the main one, so when
    affixing a reduction, one should set ``scan=True``, or ``scan=n``.

    .. marble::
        :alt: affix

         ---a1---------a2-----------a3-------|
        x----4-----------5---------6---------|
        y-----7--------8----------9----------|
        [           affix(b=x, c=y)          ]
         -----a1,b4,c7---a2,b5,c8---a3,b6,c9-|

    Example:
        .. code-block:: python

            obs.where("x", "y").affix(
                minx=obs["x"].min(scan=True),
                xpy=obs["x", "y"].starmap(lambda x, y: x + y),
            )

        Or:

        .. code-block:: python

            obs.where("x", "y").affix(
                # o is obs.where("x", "y")
                minx=lambda o: o["x"].min(scan=True),
                xpy=lambda o: o["x", "y"].starmap(lambda x, y: x + y),
            )

    Arguments:
        streams: A mapping from extra keys to add to the dicts to Observables
            that generate the values, or to functions of one argument that will
            be called with the main Observable.
    """
    from .gvn import ObservableProxy

    keys = list(streams.keys())
    values = list(streams.values())

    def create(source):
        newobs = [
            v if hasattr(v, "subscribe") else v(ObservableProxy(source)) for v in values
        ]

        def merge(elems):
            main, *rest = elems
            return {**main, **dict(zip(keys, rest))}

        return rxop.pipe(rxop.zip(*newobs), rxop.map(merge))(source)

    return create


def augment(**fns):
    """Augment a stream of dicts with new keys.

    Each key in ``fns`` should be associated to a function that will be called
    with the rest of the data as keyword arguments, so the argument names
    matter. The results overwrite the old data, if any keys are in common.

    .. note::
        The functions passed in ``fns`` will be wrapped with
        :func:`~giving.utils.lax_function` if possible.

        This means that these functions are considered to have an
        implicit ``**kwargs`` argument, so that any data they do
        not need is ignored.

    .. code-block:: python

        # [{"x": 1, "y": 2}, ...] => [{"x": 1, "y": 2, "z": 3}, ...]
        gv.augment(z=lambda x, y: x + y)

        # [{"lo": 2, "hi": 3}, ...] => [{"lo": 2, "hi": 3, "higher": 9}, ...]
        gv.augment(higher=lambda hi: hi * hi)

    Arguments:
        fns: A map from new key names to the functions to compute them.
    """

    fns = {k: lax_function(fn) for k, fn in fns.items()}
    return rxop.map(lambda data: {**{k: fn(**data) for k, fn in fns.items()}, **data})


def as_(key):
    """Make a stream of dictionaries using the given key.

    For example, ``[1, 2].as_("x")`` => ``[{"x": 1}, {"x": 2}]``

    .. marble::
        :alt: as_

        --1---2---3--|
        [   as_(x)   ]
        --x1--x2--x3-|

    Arguments:
        key: Key under which to generate each element of the stream.
    """
    return rxop.map(lambda x: {key: x})


@reducer(default_seed=(0, 0), postprocess=rxop.starmap(lambda x, sz: x / sz))
class average:
    """Produce the average of a stream of values.

    .. marble::
        :alt: average

        --1--3--5-|
        [ average() ]
        ----------3-|

    .. marble::
        :alt: average2

        ----1----3----5------|
        [ average(scan=True) ]
        ----1----2----3------|

    .. marble::
        :alt: average3

        ----1----3----5---|
        [ average(scan=2) ]
        ----1----2----4---|

    Arguments:
        scan: If True, generate the current average on every element. If a number *n*,
            generate the average on the last *n* elements.
        seed: First element of the reduction.
    """

    def reduce(self, last, add):
        x, sz = last
        return (x + add, sz + 1)

    def roll(self, last, add, drop, last_size, current_size):
        x, _ = last
        if last_size == current_size:
            return (x + add - drop, current_size)
        else:
            return (x + add, current_size)


def _average_and_variance_postprocess(sm, v2, sz):
    avg = sm / sz
    if sz >= 2:
        var = v2 / (sz - 1)
    else:
        var = None
    return (avg, var)


@reducer(
    default_seed=(0, 0, 0),
    postprocess=rxop.starmap(_average_and_variance_postprocess),
)
class average_and_variance:
    """Produce the average and variance of a stream of values.

    .. note::

        The variance for the first element is always None.

    Arguments:
        scan: If True, generate the current average+variance on every element.
            If a number *n*, generate the average+variance on the last *n* elements.
    """

    def reduce(self, last, add):
        prev_sum, prev_v2, prev_size = last
        new_size = prev_size + 1
        new_sum = prev_sum + add
        if prev_size:
            prev_mean = prev_sum / prev_size
            new_mean = new_sum / new_size
            new_v2 = prev_v2 + (add - prev_mean) * (add - new_mean)
        else:
            new_v2 = prev_v2
        return (new_sum, new_v2, new_size)

    def roll(self, last, add, drop, last_size, current_size):
        if last_size == current_size:
            prev_sum, prev_v2, prev_size = last
            new_sum = prev_sum - drop + add
            prev_mean = prev_sum / prev_size
            new_mean = new_sum / prev_size
            new_v2 = (
                prev_v2
                + (add - prev_mean) * (add - new_mean)
                - (drop - prev_mean) * (drop - new_mean)
            )
            return (new_sum, new_v2, prev_size)
        else:
            return self.reduce(last, add)


def collect_between(start, end, common=None):
    """Collect all data between the start and end keys.

    Example:
        .. code-block:: python

            with given() as gv:
                gv.collect_between("A", "Z") >> (results := [])
                give(A=1)
                give(B=2)
                give(C=3, D=4, A=5)
                give(Z=6)
                assert results == [{"A": 5, "B": 2, "C": 3, "D": 4, "Z": 6}]

    Arguments:
        start: The key that marks the beginning of the accumulation.
        end: The key that marks the end of the accumulation.
        common: A key that must be present in all data and must have
            the same value in the whole group.
    """

    def aggro(source):
        def subscribe(obs, scheduler=None):
            dicts = {}

            def on_next(value):
                if isinstance(value, dict):
                    if common is not None:
                        if common in value:
                            key = value[common]
                        else:
                            return
                    else:
                        key = None
                    if start in value:
                        dicts[key] = dict(value)
                    elif end in value:
                        current = dicts.setdefault(key, {})
                        current.update(value)
                        obs.on_next(current)
                        del dicts[key]
                    else:
                        current = dicts.setdefault(key, {})
                        current.update(value)

            return source.subscribe(on_next, obs.on_error, obs.on_completed, scheduler)

        return rx.create(subscribe)

    return aggro


@reducer(default_seed=0)
class count:
    """Count operator.

    Returns an observable sequence containing a value that represents how many elements in the specified
    observable sequence satisfy a condition if provided, else the count of items.

    Arguments:
        predicate: A function to test each element for a condition.
        scan: If True, generate a running count, if a number *n*, count the number of elements/matches
            in the last *n* elements.
    """

    def __init__(self, predicate=None):
        self.predicate = predicate

    def reduce(self, last, new):
        if self.predicate is None or self.predicate(new):
            return last + 1
        else:
            return last

    def roll(self, last, new, drop, last_size, current_size):
        if last_size == current_size:
            if self.predicate is None:
                return last
            else:
                plus = 1 if self.predicate(new) else 0
                minus = 1 if self.predicate(drop) else 0
                return last + plus - minus
        else:
            return self.reduce(last, new)


def format(string, skip_missing=False):
    """Format an object using a format string.

    Arguments:
        string: The format string.
        skip_missing: Whether to ignore KeyErrors due to missing entries in the format.
    """
    SKIP = object()

    def _fmt(x):
        if isinstance(x, dict):
            if skip_missing:
                try:
                    return string.format(**x)
                except KeyError:
                    return SKIP
            else:
                return string.format(**x)
        elif isinstance(x, (list, tuple)):
            return string.format(*x)
        else:
            return string.format(x)

    if skip_missing:
        return rxop.pipe(
            rxop.map(_fmt),
            rxop.filter(lambda x: x is not SKIP),
        )
    else:
        return rxop.map(_fmt)


def getitem(*keys, strict=False):
    """Extract one or more keys from a dictionary.

    If more than one key is given, a stream of tuples is produced.

    .. marble::
        :alt: getitem

        --x1--x2--x3-|
        [ getitem(x) ]
        --1---2---3--|

    Arguments:
        keys: Names of the keys to index with.
        strict: If true, every element in the stream is required to
            contains this key.
    """
    if len(keys) == 1:
        (key,) = keys
        if strict:
            return rxop.map(operator.itemgetter(key))
        else:
            return rxop.pipe(
                rxop.filter(lambda arg: key in arg),
                rxop.map(operator.itemgetter(key)),
            )
    else:
        if strict:
            return rxop.map(lambda arg: tuple(arg[key] for key in keys))
        else:
            return rxop.pipe(
                rxop.filter(lambda arg: builtins.all(key in arg for key in keys)),
                rxop.map(lambda arg: tuple(arg[key] for key in keys)),
            )


def group_wrap(name, **conditions):
    """Return a stream of observables for wrapped groups.

    In this schema, B and E correspond to the messages sent in the enter and exit
    phases respectively of the :meth:`~Giver.wrap` context manager.

    .. marble::
        :alt: group_wrap

        --B-a1-a2-E-a3-B-a4-E--|
        [     group_wrap()     ]
        ---+-----------+-------|
           +a1-a2-|
                       +-a4-|

    Example:
        .. code-block:: python

            results = []

            @obs.group_wrap().subscribe
            def _(obs2):
                obs2["a"].sum() >> results

    Arguments:
        name: Name of the wrap block to group on.
        conditions: Maps a key to the value it must be associated to in the
            dictionary of the wrap statement, or to a predicate function on the
            value.
    """
    from .gvn import ObservableProxy

    def oper(source):
        opens = rxop.pipe(
            where("$wrap", **conditions),
            getitem("$wrap"),
            where(step="begin", name=name),
        )

        def closes(data):
            return rxop.pipe(
                getitem("$wrap", strict=False), where(step="end", id=data["id"])
            )(source)

        return source.pipe(
            where("!$wrap"),
            rxop.window_toggle(
                openings=opens(source),
                closing_mapper=closes,
            ),
            rxop.map(ObservableProxy),
        )

    return oper


def keep(*keys, **remap):
    """Keep certain dict keys and remap others.

    .. marble::
        :alt: keep

        --x1,z2--y3--z4--|
        [   keep(x, y)   ]
        --x1-----y3------|

    Arguments:
        keys: Keys that must be kept
        remap: Keys that must be renamed
    """
    remap = {**{k: k for k in keys}, **remap}

    def _filt(data):
        return isinstance(data, dict) and any(k in remap for k in data.keys())

    def _rekey(data):
        return {k2: data[k1] for k1, k2 in remap.items() if k1 in data}

    return rxop.pipe(rxop.filter(_filt), rxop.map(_rekey))


def kfilter(fn):
    """Filter a stream of dictionaries.

    Example:

        .. code-block:: python

            # [{"x": 1, "y": 2}, {"x": 100, "y": 50}] => [{"x": 100, "y": 50}]
            gv.kfilter(lambda x, y: x > y)

    Arguments:
        fn: A function that will be called for each element, passing the
            element using ``**kwargs``.

            .. note::

                If the dict has elements that are not in the function's
                arguments list and the function does not have a ``**kwargs``
                argument, these elements will be dropped and no error will
                occur.
    """
    fn = lax_function(fn)
    return rxop.filter(lambda kwargs: fn(**kwargs))


def kmap(_fn=None, **_fns):
    """Map a dict, passing keyword arguments.

    ``kmap`` either takes a positional function argument or keyword arguments
    serving to build a new dict.

    Example:

        .. code-block:: python

            # [{"x": 1, "y": 2}] => [3]
            gv.kmap(lambda x, y: x + y)

            # [{"x": 1, "y": 2}] => [{"z": 3}]
            gv.kmap(z=lambda x, y: x + y)

    Arguments:
        _fn: A function that will be called for each element, passing the
            element using ``**kwargs``.

            .. note::

                If the dict has elements that are not in the function's
                arguments list and the function does not have a ``**kwargs``
                argument, these elements will be dropped and no error will
                occur.
        _fns: Alternatively, build a new dict with each key associated to
            a function with the same interface as fn.
    """
    if _fn and _fns or not _fn and not _fns:
        raise TypeError(
            "kmap either takes one argument or keyword arguments but not both"
        )

    elif _fn:
        _fn = lax_function(_fn)
        return rxop.map(lambda data: _fn(**data))

    else:
        fns = {k: lax_function(fn) for k, fn in _fns.items()}
        return rxop.map(lambda data: {k: fn(**data) for k, fn in fns.items()})


def kmerge(scan=False):
    """Merge the dictionaries in the stream.

    .. marble::
        :alt: kmerge

        --x1--y2--x3--z4-|
        [          kmerge()        ]
        -----------------x3,y2,z4--|

    .. marble::
        :alt: kmerge2

        --x1--y2-----x3-----z4-------|
        [      kmerge(scan=True)     ]
        --x1--x1,y2--x3,y2--x3,y2,z4-|
    """

    def _merge(last, new):
        return {**last, **new}

    if scan:
        assert isinstance(scan, bool)
        return rxop.scan(_merge)
    else:
        return rxop.reduce(_merge)


def kscan():
    """Alias for ``kmerge(scan=True)``."""
    return kmerge(scan=True)


@reducer
class min:
    """Produce the minimum of a stream of values.

    .. marble::
        :alt: minimum

        --3--2--7--6-|
        [     min()    ]
        -------------2-|

    Arguments:
        key: A key mapping function.
        comparer: A function of two elements that returns -1 if the first is smaller
            than the second, 0 if they are equal, 1 if the second is larger.
        scan: If True, generate the current minimum on every element.
        seed: First element of the reduction.
    """

    def __init__(self, key=None, comparer=None):
        self.comparer = comparer or operator.gt
        self.key = key or (lambda x: x)

    def reduce(self, last, new):
        lastc = self.key(last)
        newc = self.key(new)
        if self.comparer(lastc, newc) <= 0:
            return last
        else:
            return new


@reducer
class max:
    """Produce the maximum of a stream of values.

    .. marble::
        :alt: maximum

        --3--2--7--6-|
        [     max()    ]
        -------------7-|

    Arguments:
        key: A key mapping function.
        comparer: A function of two elements that returns -1 if the first is smaller
            than the second, 0 if they are equal, 1 if the second is larger.
        scan: If True, generate the current maximum on every element.
        seed: First element of the reduction.
    """

    def __init__(self, key=None, comparer=None):
        self.comparer = comparer or operator.gt
        self.key = key or (lambda x: x)

    def reduce(self, last, new):
        lastc = self.key(last)
        newc = self.key(new)
        if self.comparer(lastc, newc) > 0:
            return last
        else:
            return new


def sole(*, keep_key=False, exclude=[]):
    """Extract values from a stream of dicts with one entry each.

    .. marble::
        :alt: sole

        --x1--y2--y3--z4--|
        [     sole()      ]
        --1---2---3---4---|

    If, after removing keys from the exclusion set, any dict is empty
    or has a length superior to 1, that is an error.

    Arguments:
        keep_key: If True, return a (key, value) tuple, otherwise only
            return the value. Defaults to False.
        exclude: Keys to exclude.
    """

    if isinstance(exclude, str):
        exclude = [exclude]

    def extract(data):
        pairs = [(k, v) for k, v in data.items() if k not in exclude]
        if len(pairs) == 1:
            (kv,) = pairs
            return kv if keep_key else kv[1]
        elif len(pairs) == 0:
            raise Exception("The dict has no valid entries", data)
        else:
            raise Exception("The dict has more than one entry", data)

    return rxop.map(extract)


@reducer
def sum(last, new):
    return last + new


def tag(group="", field="$word", group_field="$group"):
    """Tag each dict or object with a unique word.

    If the item is a dict, do ``item[field] = <new_word>``, otherwise
    attempt to do ``setattr(item, field, <new_word>)``.

    These tags are displayed specially by the
    :meth:`~giving.gvn.ObservableProxy.display` method and they
    can be used to determine breakpoints with the
    :meth:`~giving.gvn.ObservableProxy.breakword` method.

    Arguments:
        group: An arbitrary group name that corresponds to an independent
            sequence of words. It determines the color in display.
        field: The field name in which to put the word
            (default: ``$word``).
        group_field: The field name in which to put the group
            (default: ``$group``).
    """
    try:
        import breakword as bw
    except ImportError:  # pragma: no cover
        raise ImportError(
            "Package `breakword` must be installed to use the tag() operator"
        )

    grp = bw.groups[group]

    def tag_data(data):
        word = grp.gen()
        if isinstance(data, dict):
            data = {field: word, **data}
            if group:
                data[group_field] = group
        else:
            if group:
                setattr(data, group_field, group)
                setattr(data, "$group", group)
            setattr(data, field, word)
            setattr(data, "$word", word)
        return data

    return rxop.pipe(rxop.map(tag_data), rxop.share())


def variance(*args, **kwargs):
    return rxop.pipe(
        average_and_variance(*args, **kwargs), rxop.starmap(lambda avg, var: var)
    )


def where(*keys, **conditions):
    """Filter entries with the given keys meeting the given conditions.

    .. marble::
        :alt: where

        ---a1--b2--c3--|
        [  where(b=2)  ]
        -------b2------|

    .. marble::
        :alt: where2

        ---a1--b2--a3--|
        [   where(a)   ]
        ---a1------a3--|

    Example:
        .. code-block:: python

            where("x", "!y", z=True, w=lambda x: x > 0)

    Arguments:
        keys: Keys that must be present in the dictionary or, if a key starts
            with "!", it must *not* be present.
        conditions: Maps a key to the value it must be associated to in the
            dictionary, or to a predicate function on the value.
    """
    conditions = {
        k: cond if isinstance(cond, FunctionType) else lambda x, value=cond: value == x
        for k, cond in conditions.items()
    }
    excluded = [k.lstrip("!") for k in keys if k.startswith("!")]
    keys = [k for k in keys if not k.startswith("!")]
    keys = (*keys, *conditions.keys())

    def filt(x):
        return (
            isinstance(x, dict)
            and builtins.all(k in x for k in keys)
            and not builtins.any(k in x for k in excluded)
            and builtins.all(cond(x[k]) for k, cond in conditions.items())
        )

    return rxop.filter(filt)


def where_any(*keys):
    """Filter entries with any of the given keys.

    .. marble::
        :alt: where_any

        ---a1--b2--c3--|
        [ where_any(b) ]
        -------b2------|

    Arguments:
        keys: Keys that must be present in the dictionary.
    """

    def _filt(data):
        return isinstance(data, dict) and any(k in data for k in keys)

    return rxop.filter(_filt)


__all__ = [
    "affix",
    "as_",
    "augment",
    "average",
    "average_and_variance",
    "collect_between",
    "count",
    "format",
    "getitem",
    "group_wrap",
    "keep",
    "kfilter",
    "kmap",
    "kmerge",
    "kscan",
    "max",
    "min",
    "roll",
    "sole",
    "sum",
    "tag",
    "variance",
    "where",
    "where_any",
]



import hashlib
import os
import pdb
from datetime import datetime

try:
    import breakword as bw
except ImportError:  # pragma: no cover
    bw = None


def _dethash(s):
    return int(hashlib.md5(s.encode()).hexdigest(), base=16)


def _term_color(key):
    """Generate a terminal color for this group."""
    return 30 + (_dethash(key) % 8)


_path_replacements = {
    "PWD": "",
    "CONDA_PREFIX": "$CONDA_PREFIX/",
    "VIRTUAL_ENV": "$VIRTUAL_ENV/",
    "HOME": "~/",
}


def format_libpath(path):
    for var, pfx in sorted(_path_replacements.items(), key=lambda kv: kv[1]):
        val = os.environ.get(var, None)
        if val is not None:
            if not val.endswith("/"):
                val += "/"
            if path.startswith(val):
                return os.path.join(pfx, path[len(val) :])
    else:
        return path


def extract_word(data, method="pop"):
    group = word = None

    if isinstance(data, dict):
        word = getattr(data, method)("$word", None)
        if word is not None:
            group = getattr(data, method)("$group", "")
        return group, word

    elif hasattr(data, "$word"):
        word = getattr(data, "$word", None)
        if word is not None:
            group = getattr(data, "$group", "")

    return group, word


class Displayer:
    def __init__(self, colors=True, time_format="%Y-%m-%d %H:%M:%S"):
        self.colors = colors
        self.time_format = time_format

    def with_color(self, k, color):
        if self.colors:
            return f"\033[{color};40m{k}\033[0m"
        else:
            return k

    def colorize(self, k):
        return self.with_color(k, f"1;{_term_color(k)}")

    def string(self, data):
        s = ""

        if isinstance(data, dict):
            data = dict(data)

        group, word = extract_word(data)
        if word is not None:
            word = f"{group}:{word}" if group else f"{word}"
            gcol = _term_color(group)
            s += self.with_color("«", f"0;{gcol}")
            s += self.with_color(word, f"1;{gcol}")
            s += self.with_color("» ", f"0;{gcol}")

        if isinstance(data, dict):
            time = data.pop("$time", None)
            if time is not None:
                stime = datetime.utcfromtimestamp(time).strftime(self.time_format)
                s += self.with_color(f"[{stime}] ", "1;30")

            line = data.pop("$line", None)
            if line is not None:
                fname = format_libpath(line.filename)
                s += self.colorize(f"({fname}:{line.lineno} {line.name}) ")

            entries = [f"{self.colorize(k + ':')} {v}" for k, v in data.items()]
            s += "; ".join(entries)
            return s

        else:
            return f"{s}{data}"

    def __call__(self, data):
        print(self.string(data))


display = Displayer()


class Breakpoint:  # pragma: no cover
    def __init__(self, use_breakword=False, word=None, skip=[]):
        self.word = word
        self.skip = ["giving.*", "rx.*", *skip]
        self.bw = False
        if use_breakword:
            try:
                import breakword as bw
            except ImportError:
                raise ImportError(
                    "Package `breakword` must be installed to use the tag() operator"
                )
            self.bw = bw

    def __call__(self, data):
        if self.bw:
            group, word = extract_word(data, method="get")
            if word is not None:
                results = self.bw._get_watch(group, self.word)
                if word not in results:
                    return

        print("Breaking on", display.string(data))
        try:
            breakpoint(skip=self.skip)
        except TypeError:
            pdb.Pdb(skip=self.skip).set_trace()



import functools
import types

from rx import operators as rxop
from rx.operators import NotSet


def keyword_decorator(deco):
    """Wrap a decorator to optionally takes keyword arguments."""

    @functools.wraps(deco)
    def new_deco(fn=None, **kwargs):
        if fn is None:

            @functools.wraps(deco)
            def newer_deco(fn):
                return deco(fn, **kwargs)

            return newer_deco
        else:
            return deco(fn, **kwargs)

    return new_deco


def lax_function(fn):
    """Add a ``**kwargs`` argument to fn if it does not have one.

    ``fn`` is not modified. If it has a varkw argument, ``fn`` is returned directly,
    otherwise a new function is made that takes varkw and does nothing with them (it
    just drops them)

    Example:

        .. code-block:: python

            @lax_function
            def f(x):
                return x * x

            f(x=4, y=123, z="blah")  # works, returns 16. y and z are ignored.
    """
    # NOTE: The purpose of ``lax_function`` is to simplify ``kmap``, ``kfilter``, ``ksubscribe``
    # and other mappers that can be partially applied to the element dictionaries in an
    # observable stream.

    if isinstance(fn, types.FunctionType):
        KWVAR_FLAG = 8
        co = fn.__code__
        if not co.co_flags & KWVAR_FLAG:
            if hasattr(co, "replace"):
                newco = co.replace(
                    # Let's lie and pretend it has **kwargs
                    co_flags=co.co_flags | KWVAR_FLAG,
                    # Add a dummy keyword argument with an illegal name
                    co_varnames=(*co.co_varnames, "#"),
                    # That dummy argument is a new local
                    co_nlocals=co.co_nlocals + 1,
                )
            else:  # pragma: no cover
                newco = types.CodeType(
                    co.co_argcount,
                    co.co_kwonlyargcount,
                    co.co_nlocals + 1,
                    co.co_stacksize,
                    co.co_flags | KWVAR_FLAG,
                    co.co_code,
                    co.co_consts,
                    co.co_names,
                    (*co.co_varnames, "#"),
                    co.co_filename,
                    co.co_name,
                    co.co_firstlineno,
                    co.co_lnotab,
                    co.co_freevars,
                    co.co_cellvars,
                )

            newfn = types.FunctionType(
                name=fn.__name__,
                code=newco,
                globals=fn.__globals__,
                closure=fn.__closure__,
            )
            newfn.__defaults__ = fn.__defaults__
            newfn.__kwdefaults__ = fn.__kwdefaults__
            return newfn

    return fn


class _Reducer:
    def __init__(self, reduce, roll):
        self.reduce = reduce
        self.roll = roll


@keyword_decorator
def reducer(func, default_seed=NotSet, postprocess=NotSet):
    """Create a reduction operator.

    .. note::
        If ``func`` is not given, this returns a decorator.

    * If func is a function, it should have the signature ``func(accum, x)``
      where ``accum`` is the last result and ``x`` is the new data.
    * If func is a class, it should have the following methods:

      * ``reduce(accum, x)``, used when ``scan`` is boolean (``scan=False`` or ``scan=True``)
      * ``roll(accum, x, drop, last_size, current_size)`` (see :func:`giving.operators.roll`),
        used when ``scan`` is an integer (``scan=n``). It should implement an optimized way to
        reduce over the last n elements of a sequence. Note that ``last_size`` and ``current_size``
        cannot exceed ``n``.
      * Any arguments to ``__init__`` will be passed over when the operator is called.

    Example:

    .. code-block:: python

        @reducer
        def sum(last, new):
            return last + new

    .. code-block:: python

        @reducer
        class sum:
            def reduce(self, last, new):
                return last + new

            def roll(self, last, new, drop, last_size, current_size):
                result = last + new
                if last_size == current_size:
                    result -= drop
                return result

    Arguments:
        func: A function or class that defines the reduction.
        default_seed: The default seed to start the reduction.
        postprocess: A postprocessing step to apply to the result of
            the operator.

    Returns:
        A reduction operator that takes a ``scan`` argument to perform
        a scan or roll operation instead of reduce.
    """

    from .extraops import roll

    name = func.__name__
    if isinstance(func, type):
        constructor = func

    else:

        def constructor():
            return _Reducer(reduce=func, roll=None)

    def _create(*args, scan=False, seed=NotSet, **kwargs):
        reducer = constructor(*args, **kwargs)

        if seed is NotSet:
            seed = default_seed

        if scan is True:
            oper = rxop.scan(reducer.reduce, seed=seed)

        elif scan:
            oper = roll(n=scan, reduce=reducer.roll, seed=seed)

        else:
            oper = rxop.reduce(reducer.reduce, seed=seed)

        if postprocess is not NotSet:
            oper = rxop.pipe(oper, postprocess)

        return oper

    _create.__name__ = name
    _create.__doc__ = func.__doc__
    return _create



from rx.operators import (  # noqa: F401
    all,
    amb,
    as_observable,
    buffer,
    buffer_toggle,
    buffer_when,
    buffer_with_count,
    buffer_with_time,
    buffer_with_time_or_count,
    catch,
    combine_latest,
    concat,
    contains,
    debounce,
    default_if_empty,
    delay,
    delay_subscription,
    delay_with_mapper,
    dematerialize,
    distinct,
    distinct_until_changed,
    do,
    do_action,
    do_while,
    element_at,
    element_at_or_default,
    exclusive,
    expand,
    filter,
    filter_indexed,
    finally_action,
    find,
    find_index,
    first,
    first_or_default,
    flat_map,
    flat_map_indexed,
    flat_map_latest,
    fork_join,
    group_by,
    group_by_until,
    group_join,
    ignore_elements,
    is_empty,
    join,
    last,
    last_or_default,
    map,
    map_indexed,
    materialize,
    merge,
    merge_all,
    multicast,
    observe_on,
    on_error_resume_next,
    pairwise,
    partition,
    partition_indexed,
    pipe,
    pluck,
    pluck_attr,
    publish,
    publish_value,
    reduce,
    ref_count,
    repeat,
    replay,
    retry,
    sample,
    scan,
    sequence_equal,
    share,
    single,
    single_or_default,
    single_or_default_async,
    skip,
    skip_last,
    skip_last_with_time,
    skip_until,
    skip_until_with_time,
    skip_while,
    skip_while_indexed,
    skip_with_time,
    slice,
    some,
    starmap,
    starmap_indexed,
    start_with,
    subscribe_on,
    switch_latest,
    take,
    take_last,
    take_last_buffer,
    take_last_with_time,
    take_until,
    take_until_with_time,
    take_while,
    take_while_indexed,
    take_with_time,
    throttle_first,
    throttle_with_mapper,
    throttle_with_timeout,
    time_interval,
    timeout,
    timeout_with_mapper,
    timestamp,
    to_dict,
    to_future,
    to_iterable,
    to_list,
    to_marbles,
    to_set,
    while_do,
    window,
    window_toggle,
    window_when,
    window_with_count,
    window_with_time,
    window_with_time_or_count,
    with_latest_from,
    zip,
    zip_with_iterable,
    zip_with_list,
)

from .extraops import *  # noqa: F401, F403

###########
# Aliases #
###########


norepeat = distinct_until_changed
throttle = throttle_first


__all__ = [k for k in globals().keys() if not k.startswith("_")]
